#! /bin/sh
# This is a shell archive, meaning:
# 1. Remove everything above the #! /bin/sh line.
# 2. Save the resulting text in a file.
# 3. Execute the file with /bin/sh 
export PATH; PATH=/bin:/usr/bin:$PATH
cat << \SHAR_EOF > 'auth.channel'
# AUTH.CHANNEL table not used in simple configuration
SHAR_EOF
cat << \SHAR_EOF > 'auth.mta'
# AUTH.MTA table not used in simple configuration
SHAR_EOF
cat << \SHAR_EOF > 'auth.user'
# AUTH.USER table not used in simple configuration
SHAR_EOF
cat << \SHAR_EOF > 'channel'
# A small list of top level domains...
# everything is sent to the given server by smtp
uk:_my_smtp_server_(smtp)
com:_my_smtp_server_(smtp)
edu:_my_smtp_server_(smtp)
gov:_my_smtp_server_(smtp)
mil:_my_smtp_server_(smtp)
us:_my_smtp_server_(smtp)
org:_my_smtp_server_(smtp)
SHAR_EOF
cat << \SHAR_EOF > 'domain'
# Known domains
# Our site first
_my_host_._my_site_:_my_host_._my_site_
_my_site_:_my_site_
# Some wildcards for top level domains
uk:uk *
com:com *
edu:edu *
gov:gov *
mil:mil *
us:us *
org:org *
SHAR_EOF
cat << \SHAR_EOF > 'or'
# OR table not used in simple configuration
SHAR_EOF
cat << \SHAR_EOF > 'or2rfc'
# OR2RFC table not used in simple configuration
SHAR_EOF
cat << \SHAR_EOF > 'rfc2or'
# RFC2OR table not used in simple configuration
SHAR_EOF
cat << \SHAR_EOF > 'tailor'
#
############################################################
#
# @(#) $Header: /cs/research/pp/hubris/pp-beta/config/RCS/sample.sh,v 5.0 90/09/20 16:35:14 pp Exp Locker: pp $
#
# $Log:	sample.sh,v $
# Revision 5.0  90/09/20  16:35:14  pp
# rcsforce : 5.0 public release
# 
# Revision 4.3  90/09/10  21:19:02  pp
# rcsforce : final beta
# 
#
############################################################
#
#  **-- Logging of this tailor file when it is being read --**
#
authlog level=fatal, size=400
operlog level=fatal, size=400
normlog level=fatal, size=400

#  **-- Special strings --**
# TAILOR THESE
loc_dom_site    _my_site_
loc_dom_mta     _my_host_._my_site_
# THIS NEEDS FIXING BEFORE YOU ATTEMPT X.400
loc_or          "/OU=_my_OU_/O=_my_org_/PRMD=_my_PRMD_/ADMD=_my_ADMD/C=_my_C_/"
postmaster	postmaster@_my_site_
pptsapd_addr	Internet=_my_host_+20001
qmgrhost	_my_host_
mboxname	ppmailbox
delim1		"\001\001\001\001\012"
delim2		"\001\001\001\001\012"


# END OF TAILOR THESE

pplogin         pp

authchannel	free, warnsender=warn1, warnrecipient=warn2
authloglevel	high
wrndfldir	warnings

# Body types

bodypart	hdr.822 hdr.822-jntmail hdr.822-us hdr.p2 hdr.822-uk
bodypart	ia5 p2 ipm


#  **-- Mapping tables --**
#
# All of these table MUST have the names given here
# the files can be empty initally...
tbl     aliases         show="Aliases: mapping -> local id",
                        flags=dbm
tbl     users           show="Users: mapping local id -> disposition",
                        flags=dbm
tbl     domain          show="Mapping domain key ->full domain/MTA",
                        flags=dbm
tbl     or              show="Mapping O/R Address -> MTA",flags=dbm
tbl     channel         show="Binding MTA -> Channels",flags=dbm
tbl     or2rfc          show="RFC 987: X.400 -> RFC 822",flags=dbm
tbl     rfc2or          show="RFC 987: RFC 822 -> X.400",flags=dbm
tbl	auth.channel	show="Authorisation: channel policy",flags=linear
tbl	auth.mta	show="Authorisation: mta based",flags=linear
tbl	auth.user	show="Authorisation: user based",flags=linear
#  **-- Channel tables --**
#
# Needed by individual channels, not submit
#
tbl     local           file="ch.local",
                        show="local id -> user id + home directory",
                        flags=dbm

#  **-- "Real" Channels --**
#

chan    822-local           prog=local,
        show="Local Delivery channel",type=both,
        sort="user time",adr=822,adr-order=ukpref,
        bptin="ia5,hdr.822",bptout="ia5,hdr.822-uk",
        content-out=822,table=local,access=mts,chanout=dr2rfc


# SMTP channel uses /etc/hosts or DNS
chan    smtp            prog=smtp,show="with SMTP (PP)",type=out,
                        adr=822,adr-order=usapref,
                        bptout="ia5,hdr.822-us",content-out=822,
                        chanout=dr2rfc


# To send back errors
chan    dr2rfc          prog=dr2rfc,show="Dr2rfc channel",type=out,
			adr=822,adr-order=ukpref,
			bptin="hdr.822, ia5",
			content-out=822

# QMGR special channels (load/manage/clean)
chan    qmgr-load       prog=qmgr-load,show="Loading the QMGR",
                        type=qmgrload

chan    msg-clean       prog=msg-clean,show="Removing finished message"
                        type=delete

chan    trash           prog=trash,show="Removing trash"
                        type=debris

chan    timeout		prog=timeout,show="Timeout messages"
                        type=timeout

#  **-- shaper channels --**

#   Flatteners

chan    p2flatten       prog=p2flatten,show="Structured message -> P2",
			type=shaper,content-out=p2,chanout=dr2rfc

chan    822flatten      prog=rfc934,type=shaper,content-out=822,

#   Unflatteners

chan    p2explode       prog=p2explode,type=shaper,content-in=p2,chanout=dr2rfc

# Header Filters
chan    822touk         prog=fcontrol,type=shaper,
                        bptin=hdr.822,bptout=hdr.822-uk,chanout=dr2rfc,
                        info="rfc822norm -jnt -bigend -jntsender $(822SENDER)"
chan    822tous         prog=fcontrol,type=shaper,bptin=hdr.822,
                        bptout=hdr.822-us,chanout=dr2rfc,
			info="rfc822norm -822 -littleend"
chan    822top2         prog=c-RFCtoP2,type=shaper,bptin=hdr.822,
                        bptout=hdr.p2,chanout=dr2rfc,
                        show="RFC987 RFC 822 -> P2"
chan    p2to822         prog=c-P2toRFC,type=shaper,bptout=hdr.822,
                        bptin=hdr.p2,chanout=dr2rfc,
                        show="RFC987 P2 -> RFC 822"

#  **-- Logging for the rest of the prog --**
#
authlog level=notice, size=400
operlog level=notice, size=400
normlog level=notice, size=400

822flatten      normlog  file=filter            level=notice,   size=400
p2flatten       normlog  file=filter            level=notice,   size=400
822touk         normlog  file=filter            level=notice,   size=400
822tous         normlog  file=filter            level=notice,   size=400
dr2rfc          normlog  file=filter            level=notice,   size=400
fcontrol        normlog  file=filter            level=notice,   size=400
odif2txt        normlog  file=filter            level=notice,   size=400
p2explode       normlog  file=filter            level=notice,   size=400
rfc822norm      normlog  file=filter            level=notice,   size=400
trash           normlog  file=filter            level=notice,   size=400

822top2         normlog  file=rfc987            level=notice,   size=400
p2to822         normlog  file=rfc987            level=notice,   size=400

qmgr            normlog  file=qmgr              level=notice,   size=400
qmgr-load       normlog  file=qmgr              level=notice,   size=400

smtp            normlog  file=smtp              level=pdus,   size=400
smtpsrvr        normlog  file=smtp              level=notice,   size=400
greyout         normlog  file=greybook          level=trace,    size=400
submit          normlog  file=submit            level=notice,   size=400

local		normlog	 file=local		level=notice,	size=400
msg-clean	normlog	 file=msg-clean		level=notice,	size=400

x400in84        normlog  file=x40084            level=notice,   size=400
x400out84       normlog  file=x40084            level=notice,   size=400

isodelog psaplevel dlevel=all level=exceptions level=fatal level=notice
SHAR_EOF
exit 0
#	End of shell archive
